import os
import socket
import asyncio
import pickle
import json
import logging
from collections import defaultdict
import nacos
import psutil
from copy import deepcopy
import uuid
from fastapi import FastAPI, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
import uvicorn
import time

from data_schemas import *
from grammar_parser import init_object, GrammarBasedParser
from dialog_management import DialogStatus
from recall_query import QueryRecall
from model_hub import ModelHub
from anomaly_detection import AnomalyDetection
import constants
from cut_tool import cut_tool
from logging_util import logger, LOG_PATH
from exception_handler import *
from helper_funcs import remove_spaces, parse_schema, LRUDictionary, count_mention_hit, select_model, \
    sort_dict_by_key_length, replace_mention_by_disambiguation, domain_question_handle, filter_question, modify_intent_by_reject_fields, filter_enum
import pandas as pd
from datetime import datetime
from intent import Field
import numpy as np
from conf import *
from redis_util import RedisUtils
from database_helper import RedisHelper
from nacos_helpers import register_nacos_service, deregister_nacos_service
from utils import PerformanceTimer


lock_query_to_intent = asyncio.Lock()
lock_query_cut = asyncio.Lock()

INVALID_API_KEY = {"successful": False, "message": "Invalid api key"}


def not_found_parser(subject_id):
    return {
        "successful":
            False,
        "message":
            f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/"
    }

def not_found_session(session_id):
    return {
        "successful": False,
        "message": f"Found no Parser for subject_id '{session_id}'"
    }


def failed_to_parse(query, subject_id):
    return {"successful": False, "message": f"Parser internal error. Failed to parse query '{query}' in table '{subject_id}'"}


def empty_intent(query, subject_id):
    return {"successful": True, "message": f"Empty intent. Failed to parse query '{query}' in table '{subject_id}'"}


def failed_to_init_parser(subject_id):
    return {"successful": False, "message": f"Failed to initialize Parser for table '{subject_id}'"}


def create_parser(schema_dict):
    logger.info(f"--------------------------------------NL_object初始化{schema_dict['model_id']}-----------------------------------------")
    nl_object = init_object(schema_dict=schema_dict)
    logger.info("--------------------------------------NL_object初始化结束--------------------------------------")
    logger.info("-----------------------------------------------------------------------------------------")
    logger.info(f"----------------------------------------Parser初始化{schema_dict['model_id']}----------------------------------------")
    parser = GrammarBasedParser(nl_object)
    logger.info("--------------------------------------Parser初始化结束--------------------------------------")
    logger.info("------------------------------------------------------------------------------------------------------")
    return parser

def create_query_util():
    logger.info("--------------------------------------query_util初始化-----------------------------------------")
    query_recall = QueryRecall()
    logger.info("--------------------------------------query_util初始化结束-----------------------------------------")
    return query_recall


def check_and_create_dir(dir_path_list):
    """检查文件夹是否存在，否则新建一个文件夹"""
    for dir_path in dir_path_list:
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
            logger.info(f"Not found {dir_path}, created a new one")


def all_files(root):
    for root, dir_list, file_list in os.walk(root):
        for file in file_list:
            yield os.path.join(root, file)


def save_schema_in_pkl_file(config_body: ConfigBody, file_dir: str):
    if not file_dir.endswith(('/', '\\')):
        file_dir = file_dir + '/'
    with open(f'{file_dir}{config_body.subject_id}.pkl', 'wb') as f:
        pickle.dump(config_body.dict(), f)

def save_schema(config_body: ConfigBody):
    del config_body.api_key
    save_schema_in_pkl_file(config_body, TABLE_SCHEMA_DIR)

def save_persist_subject(subject_body: SubjectBody):
    disk_files = os.listdir(TABLE_SCHEMA_DIR)
    disk_files_name = [_file.replace(".pkl", "") for _file in disk_files]
    if subject_body.subject_id in disk_files_name:
        delete_file_path = TABLE_SCHEMA_DIR + os.path.sep + subject_body.subject_id + ".pkl"
        os.remove(delete_file_path)
    save_path = TABLE_SCHEMA_DIR + os.path.sep + subject_body.subject_id + ".pkl"
    with open(save_path, "wb") as f:
        pickle.dump(subject_body.dict(), f)

def save_temp_subject(subject_body: SubjectBody):
    disk_files = os.listdir(TEMP_TABLE_SCHEMA_DIR)
    disk_files_name = [_file.replace(".pkl", "") for _file in disk_files]
    if subject_body.subject_id in disk_files_name:
        delete_file_path = TEMP_TABLE_SCHEMA_DIR + os.path.sep + subject_body.subject_id + ".pkl"
        os.remove(delete_file_path)
    save_path = TEMP_TABLE_SCHEMA_DIR + os.path.sep + subject_body.subject_id + ".pkl"
    with open(save_path, "wb") as f:
        pickle.dump(subject_body.dict(), f)


def truncate_query(sentence_body: SentenceBody) -> SentenceBody:
    """query 超过长度限制将被截取"""
    max_query_len = 150
    if len(sentence_body.query) > max_query_len:
        logger.info("query 超过长度限制，将被截取")
        sentence_body.query = sentence_body.query[:max_query_len]
        sentence_body.tokens = [TokenBody(
            text=sentence_body.query,
            type="ORIGIN",
            startIndex=0,
            endIndex=len(sentence_body.query),
            editable=True,
            similarList=[],
            tableField=None,
            tableId=None
        )]
    return sentence_body


def get_slang(subject):
    slang = {}
    if subject.slang_list:
        for slang_info in subject.slang_list:
            for slang_mention in slang_info.slangs:
                if not slang_mention or slang_mention.isspace(): continue
                slang[slang_mention] = slang_info.standard_mention.query
        slang = sort_dict_by_key_length(slang)
    return slang


def get_parameter_info(subject):
    model_id2parameter_alias_dict = defaultdict(dict)
    if subject.param_list:
        for parameter in subject.param_list:
            for field in parameter.fields:
                parameter_dict = {"key": parameter.key, "field": field}
                for alias in parameter.alias:
                    if not alias or alias.isspace(): continue
                    model_id2parameter_alias_dict[field.modelId][alias] = parameter_dict
    model_id2parameter_alias_dict = dict(model_id2parameter_alias_dict)
    for model_id, parameter_alias_dict in model_id2parameter_alias_dict.items():
        model_id2parameter_alias_dict[model_id] = sort_dict_by_key_length(parameter_alias_dict)
    return model_id2parameter_alias_dict


def get_model_infos(model_dict_list, model_id2parameter_alias_dict, slang):
    model_infos = []
    for model_dict in model_dict_list:
        model_id = model_dict.model_id
        attr_ac_tree, attr_mention_mapping, fuzz_dict, enum_ac_tree, metric_info, dimension_info, date_info, date_dim, add_word_set, table_id_map_to_model_id, \
            table_name2table_id, met_lower_alias_map, dim_lower_alias_map, table_name_info, enum_aliases_info = parse_schema(model_dict.table_schemas, model_id, slang)
        metric_info = sort_dict_by_key_length(metric_info)
        dimension_info = sort_dict_by_key_length(dimension_info)
        date_info = sort_dict_by_key_length(date_info)
        parser = create_parser(model_dict.dict())
        model_infos.append(
            {
                "model_id": model_id,
                "metric_info": metric_info,
                "dimension_info": dimension_info,
                "date_info": date_info,
                "date_dim": date_dim,
                "parser": parser,
                "table_id2model_id": table_id_map_to_model_id,
                "table_name2table_id": table_name2table_id,
                "met_lower_alias_map": met_lower_alias_map,
                "dim_lower_alias_map": dim_lower_alias_map,
                "table_name_info": table_name_info,
                "slang": slang,
                "enum_aliases_info": enum_aliases_info,
                "parameter_alias_dict": model_id2parameter_alias_dict.get(model_id, {}),
                "attr_mention_mapping": attr_mention_mapping,
                "attr_ac_tree": attr_ac_tree,
                "fuzz_dict": fuzz_dict,
                "enum_ac_tree": enum_ac_tree
            }
        )
    return model_infos


def init_parsers_on_start_up():
    for file_path in all_files(TABLE_SCHEMA_DIR):
        if not file_path.endswith(".pkl"):
            continue
        with open(file_path, 'rb') as f:
            subject = pickle.load(f)
        subject = SubjectBody(**subject)
        subject_id = subject.subject_id
        model_dict_list = subject.model_dict_list
        subject_name = subject.subject_name
        slang = get_slang(subject)
        model_id2parameter_alias_dict = get_parameter_info(subject)
        model_infos = get_model_infos(model_dict_list, model_id2parameter_alias_dict, slang)
        persist_table_parser_map[subject_id] = {
            "subject_name": subject_name,
            "model_infos": model_infos
        }
        logger.info(f"预加载主题id：{subject_id}")


def valid_api_key(api_key):
    return api_key in authorized_api_keys

authorized_api_keys = AUTHORIZED_API_KEYS

app = FastAPI()

nacos_client: nacos.NacosClient = None


@app.on_event("startup")
async def init_utils():
    logger.info(f"------------------------------- Version {VERSION} ----------------------------------")
    logger.info("--------------------------------------信息预加载--------------------------------------")
    check_and_create_dir([TABLE_SCHEMA_DIR, TEMP_TABLE_SCHEMA_DIR, LOG_PATH])
    global query_recall, model_hub, redis_utils, dialog_status, \
        table_parser_map, persist_table_parser_map, session_id2info

    redis_utils = None
    table_parser_map = LRUDictionary()
    persist_table_parser_map = dict()
    session_id2info = dict()

    if REDIS_ENABLE:
        with RedisStartupContext():
            redis_utils = RedisUtils(REDIS_DB)
            await redis_utils.client.ping()
        dialog_status = DialogStatus(redis_utils=redis_utils)
        logger.info("Redis 连接成功")
    else:
        # 单机状态
        logger.info("Redis 未启用，不支持多节点部署")
        dialog_status = DialogStatus()
        init_parsers_on_start_up()

    query_recall = create_query_util()
    model_hub = ModelHub()
    logger.info("--------------------------------------预加载完毕--------------------------------------")

    if NACOS_ENABLE:
        logger.info(f"已启用 NACOS 服务发现")
        global nacos_client
        with NacosContext():
            nacos_client = register_nacos_service()
    else:
        logger.info(f"NACOS 未启用，不支持服务发现")
    logger.info(f"------------------------------- Version {VERSION} ----------------------------------")


@app.on_event("shutdown")
async def shutdown_service():
    try:
        if REDIS_ENABLE:
            await redis_utils.client.aclose()
        if NACOS_ENABLE:
            deregister_nacos_service(nacos_client)
    except redis.exceptions.RedisError:
        logger.warning(f"关闭 Redis 连接失败")
    except nacos.NacosException:
        logger.warning(f"注销 Nacos 服务失败")
    except Exception as e:
        logger.exception(e)
        logger.warning(f"服务关闭时出现未知错误")
    logger.info(f"服务已关闭")


@app.middleware("http")
async def handle_response(request, call_next):
    start_time = time.perf_counter()
    response = await call_next(request)
    response.headers["Content-Type"] = "application/json; charset=UTF-8"
    logger.info(
        f"code: {response.status_code} | "
        f"URL路径: {request.url.path} | "
        f"耗时: {time.perf_counter() - start_time:.3f}s | "
        f"request_uuid: {request.headers.get('request_uuid', None)}"
    )
    return response


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request: Request, exc: RequestValidationError):
    logger.warning(f"RequestValidationError: {exc.errors()}")
    return JSONResponse(
        status_code=422,
        content={"detail": exc.errors()},
    )


@app.put("/api/v1/preload")
async def preload_parser(preload_body: PreloadBody):
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------预加载请求开始--------------------------------------")
    logger.info("******************************************************************************************")

    cur_subject_id = ""
    try:
        if not REDIS_ENABLE:
            # 单机状态下需要判断服务器内存
            memory_usage = psutil.virtual_memory()
            memory_percent = memory_usage.percent
            if memory_percent > 95:
                return {"successful": False, "message": "OOM！！！"}
            if not valid_api_key(preload_body.api_key):
                logger.error("服务器API_KEY错误！！！")
                return INVALID_API_KEY

        for subject in preload_body.datas:
            subject_id = subject.subject_id
            if not REDIS_ENABLE:
                if subject_id in persist_table_parser_map:
                    del persist_table_parser_map[subject_id]
                    delete_path = TABLE_SCHEMA_DIR + os.path.sep + subject_id + ".pkl"
                    os.remove(delete_path)
            cur_subject_id = subject_id
            model_dict_list = subject.model_dict_list
            subject_name = subject.subject_name
            slang = get_slang(subject)
            model_id2parameter_alias_dict = get_parameter_info(subject)
            model_infos = get_model_infos(model_dict_list, model_id2parameter_alias_dict, slang)
            _value = {
                "subject_name": subject_name,
                "model_infos": model_infos
            }
            await RedisHelper.save_preload_schema(
                persist_table_parser_map=persist_table_parser_map,
                redis_utils=redis_utils,
                key=subject_id,
                value=_value,
                subject=subject
            )
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.error(f"错误的subject_id：{cur_subject_id}")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return {"successful": False, "message": f"{cur_subject_id} something error"}
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------预加载请求结束--------------------------------------")
    logger.info("******************************************************************************************")
    return {"successful": True, "message": "Created a new Parser"}


@app.put("/api/v1/load")
async def load_parser(load_body: LoadBody):
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------加载请求开始--------------------------------------")
    logger.info("******************************************************************************************")

    try:
        if not REDIS_ENABLE:
            memory_usage = psutil.virtual_memory()
            memory_percent = memory_usage.percent
            if memory_percent > 95:
                logger.info("！！！！！！！！！！！！！！！！！！！！！！！OOM！！！！！！！！！！！！！！！！！！！！！！！！")
                return {"successful": False, "message": "OOM！！！"}
            if not valid_api_key(load_body.api_key):
                logger.error("服务器API_KEY错误！！！")
                return INVALID_API_KEY

        subject = load_body.data
        subject_id = subject.subject_id
        model_dict_list = subject.model_dict_list
        subject_name = subject.subject_name
        slang = get_slang(subject)
        model_id2parameter_alias_dict = get_parameter_info(subject)
        model_infos = get_model_infos(model_dict_list, model_id2parameter_alias_dict, slang)
        _value = {
            "subject_name": subject_name,
            "model_infos": model_infos
        }

        await RedisHelper.save_load_schema(
            table_parser_map=table_parser_map,
            redis_utils=redis_utils,
            key=subject_id,
            value=_value,
            subject=subject
        )

    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.error(f"错误的subject_id：{load_body.data.subject_id}")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return {"successful": False, "message": f"{load_body.data.subject_id} something error"}
    logger.info("******************************************************************************************")
    logger.info(f"------------------ 加载请求结束 -------------------------")
    logger.info(f"------------------ subject id {load_body.data.subject_id} -------------------------")
    logger.info(f"------------------ subject name {load_body.data.subject_name} -------------------------")
    logger.info("******************************************************************************************")
    return {"successful": True, "message": "Created a new Parser"}


@app.delete("/api/v1/preload/delete")
async def preload_delete_parser(delete_subject_body: DeleteSubjectBody):
    logger.info("******************************************************************************************")
    logger.info(f"--------------------------------------删除预加载主题{delete_subject_body.subject_ids}--------------------------------------")
    logger.info("******************************************************************************************")
    try:
        await RedisHelper.delete_preload_schema(
            persist_table_parser_map=persist_table_parser_map,
            subject_ids=delete_subject_body.subject_ids,
            redis_utils=redis_utils
        )
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.error(f"错误的subject_id：{delete_subject_body.subject_ids}")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return {"successful": False, "message": "something error"}
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------删除预加载主题结束--------------------------------------")
    logger.info("******************************************************************************************")
    return {"successful": True, "message": "delete"}


@app.post("/api/v1/create/session")
async def create_session(create_session_body: CreateSessionBody):
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------创建会话请求开始--------------------------------------")
    logger.info("******************************************************************************************")
    session_id = str(uuid.uuid4())
    query = create_session_body.query
    msg = ""
    result = {}
    session_info = {}
    try:
        if create_session_body.subject_ids:
            if not create_session_body.subject_ids:
                result = {
                    "session_id": None,
                    "subject_id": None
                }
            if len(create_session_body.subject_ids) == 1:
                result = {
                    "session_id": session_id,
                    "subject_id": create_session_body.subject_ids[0]
                }
                session_info["subject_id"] = create_session_body.subject_ids[0]
            else:
                tar_subject_id = ""
                tar_model_id = ""
                max_count = 0
                for subject_id in create_session_body.subject_ids:
                    model_infos = await RedisHelper.get_model_info_by_subject_id(
                        subject_id=subject_id,
                        persist_table_parser_map=persist_table_parser_map,
                        table_parser_map=table_parser_map,
                        redis_utils=redis_utils
                    )
                    if not model_infos:
                        continue
                    for model_info in model_infos:
                        model_id = model_info["model_id"]
                        res_count = count_mention_hit(query, model_info)
                        cur_count = res_count["count"]
                        if cur_count > max_count:
                            max_count = cur_count
                            tar_model_id = model_id
                            tar_subject_id = subject_id
                if max_count == 0:
                    result = {
                        "session_id": session_id,
                        "subject_id": create_session_body.subject_ids[0]
                    }
                    session_info["subject_id"] = create_session_body.subject_ids[0]
                else:
                    result = {
                        "session_id": session_id,
                        "subject_id": tar_subject_id
                    }
                    session_info["subject_id"] = tar_subject_id

        else:
            tar_subject_id = ""
            tar_model_id = ""
            max_count = 0
            exist_parser_map = await RedisHelper.get_all_preload_schemas(
                persist_table_parser_map=persist_table_parser_map,
                redis_utils=redis_utils
            )
            for subject_id, subject in exist_parser_map.items():
                model_infos = subject["model_infos"]
                for model_info in model_infos:
                    model_id = model_info["model_id"]
                    res_count = count_mention_hit(query, model_info)
                    cur_count = res_count["count"]
                    if cur_count > max_count:
                        max_count = cur_count
                        tar_model_id = model_id
                        tar_subject_id = subject_id
            if max_count == 0:
                result = {
                    "session_id": session_id,
                    "subject_id": create_session_body.subject_ids[0]
                }
                session_info["subject_id"] = create_session_body.subject_ids[0]
            else:
                result = {
                    "session_id": session_id,
                    "subject_id": tar_subject_id
                }
                session_info["subject_id"] = tar_subject_id

        session_info["param_values"] = create_session_body.param_values
        await RedisHelper.save_session_info(
            session_id=session_id,
            session_info=session_info,
            redis_utils=redis_utils,
            session_id2info=session_id2info
        )

    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return {"successful": False, "message": "something error", "results": result}
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------创建会话请求结束--------------------------------------")
    logger.info("******************************************************************************************")
    if not result:
        return {"successful": True, "results": result, "message": "subject not found"}
    return {"successful": True, "results": result, "message": ""}


@app.delete("/api/v1/close/session")
async def close_session(close_session_body: CloseSessionBody):
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------关闭会话请求--------------------------------------")
    logger.info("******************************************************************************************")
    try:
        session_id = close_session_body.session_id

        delete_flag = await RedisHelper.delete_session(
            session_id=session_id,
            redis_utils=redis_utils,
            session_id2info=session_id2info
        )

        if not delete_flag:
            return {"successful": False, "message": f"{session_id} not found"}

        await RedisHelper.delete_dialog_session(
            session_id=session_id,
            dialog_status=dialog_status,
            redis_utils=redis_utils
        )

    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.error(f"错误的session_id：{close_session_body.session_id}")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return {"successful": False, "message": "something error"}
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------关闭会话请求结束--------------------------------------")
    logger.info("******************************************************************************************")
    return {"successful": True, "message": "ok"}

@app.post("/api/v1/cut")
async def query_cut(cut_body: CutBody):
    try:
        cut_start_time = time.perf_counter()
        logger.info("******************************************************************************************")
        logger.info("--------------------------------------cut请求开始--------------------------------------")
        logger.info("*******************************************************************************************")
        logger.info(f"Query:{cut_body.sentence.query}")
        sentence = []
        result = {}

        # 锁外异步执行 io 阻塞操作
        with PerformanceTimer() as redis_timer:
            model_infos = await RedisHelper.get_model_info_by_subject_id(
                subject_id=cut_body.sentence.subjectId,
                persist_table_parser_map=persist_table_parser_map,
                table_parser_map=table_parser_map,
                redis_utils=redis_utils
            )
            pre_status_lst = await RedisHelper.get_pre_status_by_session_id(
                session_id=cut_body.session_id,
                dialog_status=dialog_status
            )

            pre_status = None if not pre_status_lst else pre_status_lst[0]

        was_cut = False
        lock_start_time = time.perf_counter()
        async with lock_query_cut:
            lock_time = time.perf_counter() - lock_start_time
            query_cut_body = cut_body.sentence
            disambiguation_body_list = cut_body.disambiguation_list if cut_body.disambiguation_list else []
            subject_id = query_cut_body.subjectId

            if not model_infos:
                logger.error(f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/")
                return not_found_parser(subject_id)

            query_cut_body = truncate_query(query_cut_body)

            with PerformanceTimer() as cut_timer:
                if not query_cut_body.editable:
                    # 不管有没有origin，都走切词逻辑
                    model_id = query_cut_body.modelId
                    model_info = None
                    for _model in model_infos:
                        if _model["model_id"] == model_id:
                            model_info = _model
                            break
                    if model_info:
                        was_cut = True
                        sentence = cut_tool.cut_query_single_model(model_info, query_cut_body.tokens, query_cut_body.query, disambiguation_body_list)
                        result["subjectId"] = subject_id
                        result["modelId"] = model_id
                        result["tokens"] = sentence
                        result["query"] = query_cut_body.query
                        result["editable"] = False
                    else:
                        was_cut = False
                        query_cut_body.tokens = []
                        logger.warning(f"query_cut_body.editable 为 False，但 modelId「{query_cut_body.modelId}」不存在。默认走多模型切词")

                if not was_cut:
                    pre_model_id = ""
                    if pre_status:
                        pre_model_id = pre_status["model_id"]

                    sentence, model_id = cut_tool.cut_query_multi_model(model_infos, query_cut_body.tokens, query_cut_body.query, disambiguation_body_list, pre_model_id)
                    result["subjectId"] = subject_id
                    result["modelId"] = model_id
                    result["tokens"] = sentence
                    result["query"] = query_cut_body.query
                    result["editable"] = True
        logger.info("******************************************************************************************")
        logger.info("--------------------------------------cut请求结束--------------------------------------")
        logger.info("******************************************************************************************")
        logger.info(f"Result: {' # '.join([token['text'] for token in sentence])}")
        logger.info(f"Type: {' # '.join([token['type'] for token in sentence])}")
        logger.info(f"UNK: {' # '.join([token['text'] for token in sentence if token['type'].startswith('UNK')])}")
        logger.info(f"**切词耗时 {time.perf_counter() - cut_start_time:.3f} (cut {cut_timer.elapsed_time:.3f}, redis {redis_timer.elapsed_time:.3f}, 锁 {lock_time:.3f})")
        return {"successful": True, "result": result}
    except Exception as e:
        result = None
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return {"successful": False, "result": result}


def replace_parameter_tokens(model_info, query_body, session_info):
    if not session_info:
        return query_body.sentence.tokens, False

    parameter_values = session_info["param_values"]
    parameter_alias_dict = model_info["parameter_alias_dict"]
    new_tokens = []
    cur_index = 0
    parameter_flag = False
    for _token in query_body.sentence.tokens:
        if _token.type != "PARAMETER":
            _token.startIndex = cur_index
            _token.endIndex = cur_index + len(_token.text) - 1
            cur_index += len(_token.text)
            new_tokens.append(_token)
        elif _token.text in parameter_alias_dict:
            parameter_flag = True
            parameter_field: ParameterField = parameter_alias_dict[_token.text]["field"]
            dimension_enum_list = parameter_values.get(parameter_alias_dict[_token.text]["key"], [])
            if not dimension_enum_list:
                logger.info(f"存在 Parameter {parameter_alias_dict[_token.text]['key']} 但对应的枚举值为空")
                _token.startIndex = cur_index
                _token.endIndex = cur_index + len(_token.text) - 1
                cur_index += len(_token.text)
                new_tokens.append(_token)
                continue
            if isinstance(dimension_enum_list, str):
                dimension_enum_list = [dimension_enum_list]
            for idx, dimension_enum in enumerate(dimension_enum_list):
                new_tokens.append(TokenBody(
                    text=dimension_enum,
                    type="DIMENSION_ENUM",
                    startIndex=cur_index,
                    endIndex=cur_index + len(dimension_enum) - 1,
                    tableId=parameter_field.tableId,
                    tableField=parameter_field.fieldName,
                    editable=False,
                    similarList=[]
                ))
                cur_index += len(dimension_enum)
                if idx < len(dimension_enum_list) - 1:
                    new_tokens.append(TokenBody(
                        text="、",
                        type="UNK",
                        startIndex=cur_index,
                        endIndex=cur_index,
                        tableId="",
                        tableField="",
                        editable=True,
                        similarList=[]
                    ))
                    cur_index += 1
        else:
            logger.error("！！！！！！！！！！！！！！个性化问法没有匹配到！！！！！！！！！！！！！！")
            logger.info(f"token: {_token.text}")
            logger.info(f"model param_values: {parameter_alias_dict}")
    return new_tokens, parameter_flag


@app.post("/api/v1/intent")
async def query_to_intent(query_body: QueryBody):
    subject_id = ""
    model_id = ""
    try:
        parse_start_time = time.perf_counter()
        logger.info("******************************************************************************************")
        logger.info("--------------------------------------intent请求开始--------------------------------------")
        logger.info("******************************************************************************************")
        if not valid_api_key(query_body.api_key):
            logger.error("服务器API_KEY错误！！！")
            return INVALID_API_KEY

        # 先选择合适的model
        subject_id = query_body.sentence.subjectId
        table_id2table_name = {} # 转换一下tableId到table_name
        table_id2model_id = {}

        # 拿到上一轮的模型id和表的id，如果走多模型切词需要判断，如果没有有效信息则走上一轮的模型，如果有similarlist的情况也优先上一轮的模型
        parser: GrammarBasedParser = None
        pre_model_id = ""

        # 锁外异步执行 io 阻塞操作
        with PerformanceTimer() as redis_timer1:
            model_infos = await RedisHelper.get_model_info_by_subject_id(
                subject_id=subject_id,
                persist_table_parser_map=persist_table_parser_map,
                table_parser_map=table_parser_map,
                redis_utils=redis_utils
            )
            if not model_infos:
                logger.error(f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/")
                return not_found_parser(subject_id)

            pre_status_lst = await RedisHelper.get_pre_status_by_session_id(
                session_id=query_body.session_id,
                dialog_status=dialog_status
            )
            session_info = await RedisHelper.get_session_info_by_session_id(
                session_id=query_body.session_id,
                redis_utils=redis_utils,
                session_id2info=session_id2info
            )

            pre_status = None if not pre_status_lst else pre_status_lst[0]

        lock_start_time = time.perf_counter()
        async with lock_query_to_intent:
            lock_time = time.perf_counter() - lock_start_time
            if pre_status:
                pre_model_id = pre_status["model_id"]

            with PerformanceTimer() as cut_timer:
                if not query_body.sentence.modelId:
                    sentence, model_id = cut_tool.cut_query_multi_model(model_infos, query_body.sentence.tokens, query_body.sentence.query, query_body.disambiguation_body_list, pre_model_id)
                    query_body.sentence.tokens = [TokenBody(**token) for token in sentence]
                else:
                    model_id = query_body.sentence.modelId

                model_info = {}

                for _model in model_infos:
                    if _model["model_id"] == model_id:
                        if not parser:
                            parser = _model["parser"]
                        model_info = _model
                        table_id2model_id = _model["table_id2model_id"]
                        for name, id in _model["table_name2table_id"].items():
                            table_id2table_name[id] = name
                        break
                logger.info(f"***************************************选择的subject_id是{subject_id}***************************************************")
                logger.info(f"***************************************选择的model_id是{model_id}***************************************************")

                if not parser:
                    logger.error(f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/")
                    return not_found_parser(subject_id)

                intent_flag = "intent_pass"

                query_body.sentence = truncate_query(query_body.sentence)

                if not query_body.sentence.tokens or not all(_token.type != "ORIGIN" for _token in query_body.sentence.tokens):
                    sentence = cut_tool.cut_query_single_model(model_info, query_body.sentence.tokens, query_body.sentence.query, query_body.disambiguation_body_list)
                    query_body.sentence.tokens = [TokenBody(**token) for token in sentence]
                    logger.info("------------------------------存在 ORIGIN token 调用切词-----------------------------------")

                slangs = model_info["slang"]
                slang_flag = False
                for _token in query_body.sentence.tokens:
                    if _token.type == "SLANG":
                        if _token.text in slangs:
                            slang_flag = True
                            _token.text = slangs[_token.text]
                            _token.type = "ORIGIN"
                            continue
                        else:
                            logger.info("！！！！！！！！！！！！！！俚语没有匹配到！！！！！！！！！！！！！！")

                query_body.sentence, filter_flag = filter_enum(query_body.sentence, model_info["enum_aliases_info"])
                if slang_flag or filter_flag:
                    logger.info(f"SLANG替换后: {' # '.join([token.text for token in query_body.sentence.tokens])}")
                    logger.info(f"Type: {' # '.join([token.type for token in query_body.sentence.tokens])}")
                    query_body.sentence.query = "".join([_t.text for _t in query_body.sentence.tokens])

                if not query_body.sentence.tokens or not all(_token.type != "ORIGIN" for _token in query_body.sentence.tokens):
                    # enable_slang 设为 False，此处切词不再有俚语
                    sentence, model_id_re_cut = cut_tool.cut_query_multi_model(model_infos, query_body.sentence.tokens, query_body.sentence.query, query_body.disambiguation_body_list, pre_model_id, enable_slang=False)
                    if model_id_re_cut != model_id:
                        query_body.sentence.modelId = model_id_re_cut
                        table_id2table_name = {}
                        for _model in model_infos:
                            if _model["model_id"] == model_id_re_cut:
                                parser = _model["parser"]
                                model_info = _model
                                table_id2model_id = _model["table_id2model_id"]
                                for name, id in _model["table_name2table_id"].items():
                                    table_id2table_name[id] = name
                                break
                        logger.info(f"***************************************经过了俚语，选择的subject_id是{subject_id}***************************************************")
                        logger.info(f"***************************************经过了俚语，选择的model_id是{model_id_re_cut}***************************************************")
                    model_id = model_id_re_cut
                    query_body.sentence.tokens = [TokenBody(**token) for token in sentence]
                    logger.info("------------------------------存在 ORIGIN token 调用切词-----------------------------------")

                # 判断是否要做多轮，如果当前model_id和之前的model_id不一致则不进多轮
                if query_body.inherited_state:
                    if pre_status:
                        if model_id != pre_status["model_id"]:
                            query_body.inherited_state = False

                parameter_replaced_tokens, parameter_flag = replace_parameter_tokens(model_info, query_body, session_info)

                if parameter_flag:
                    logger.info(f"Parameter替换后: {' # '.join([token.text for token in query_body.sentence.tokens])}")
                    logger.info(f"Type: {' # '.join([token.type for token in query_body.sentence.tokens])}")
                    query_body.sentence.tokens = parameter_replaced_tokens
                    query_body.sentence.query = "".join([_t.text for _t in query_body.sentence.tokens])

            logger.info("******************************************************************************************")
            logger.info("--------------------------------------当前切词结果--------------------------------------")
            logger.info(f"Result: {' # '.join([token.text for token in query_body.sentence.tokens])}")
            logger.info(f"Type: {' # '.join([token.type for token in query_body.sentence.tokens])}")
            logger.info(f"UNK: {' # '.join([token.text for token in query_body.sentence.tokens if token.type.startswith('UNK')])}")

            cut_tokens = query_body.sentence.tokens

            if not (''.join([token.text for token in query_body.sentence.tokens]) == query_body.sentence.query):
                logger.error("切词结果和 query 不一致！")
                logger.info(f"切词结果：{''.join([token.text for token in query_body.sentence.tokens])}")
                logger.info(f"原query：{''.join(query_body.sentence.query)}")

            for token in query_body.sentence.tokens:
                if token.similarList and token.type != "UNK":
                    token.similarList = [e for e in token.similarList if e["tableId"] in table_id2model_id]
                    token.similarList = [TokenBody(**e) for e in token.similarList if table_id2model_id[e["tableId"]] == model_id]
                    for _sim in token.similarList:
                        _sim.tableId = table_id2table_name[_sim.tableId]
                if token.tableId and token.tableId in table_id2table_name:
                    token.tableId = table_id2table_name[token.tableId]

            if query_body.disambiguation_body_list:
                # 根据modelId过滤消歧信息
                query_body.disambiguation_body_list = [_body for _body in query_body.disambiguation_body_list if _body.ignored or _body.tableId in table_id2table_name]
                for disambiguation_body in query_body.disambiguation_body_list:
                    if disambiguation_body.tableId:
                        disambiguation_body.tableId = table_id2table_name[disambiguation_body.tableId]

            # 根据消歧的信息去过滤cut_tokens
            cut_tokens = replace_mention_by_disambiguation(query_body.disambiguation_body_list, cut_tokens)

            query = remove_spaces(query_body.sentence.query)

            if all(token.type not in ["DIMENSION", "METRIC", "DIMENSION_ENUM", "TABLE_NAME", "DATE_DIMENSION"] for token in cut_tokens):
                # 只有没有匹配到任何表信息才判断要不要拒识
                reject_flag = False
                need_reject_model = True
                for regex in constants.rejection_regex_black_list:
                    if regex.search(query):
                        reject_flag = True
                for regex in constants.rejection_regex_white_list:
                    if regex.search(query):
                        need_reject_model = False

                if need_reject_model and not reject_flag:
                    intent_rejection_result = model_hub.intent_reject_model(query)[0]
                    logger.info("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
                    logger.info(intent_rejection_result["label"])
                    logger.info(intent_rejection_result["score"])
                    logger.info("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
                    if intent_rejection_result["label"] == "reject" and intent_rejection_result["score"] > 0.97:
                        reject_flag = True

                if reject_flag:
                    intent_flag = "intent_rejection"
                    logger.info("--------------------------------------intent_rejection-----------------------------------------")
                    return {"successful": True, "result": {"intent": None, "status": intent_flag}}

            # TODO 对model_hub进一步封装，将意图识别归到更上层的结构，现阶段只用触发词进归因，后期换模型+触发词的形式

            cur_dst = None
            with PerformanceTimer() as intent_timer:
                if any(_keyword in query for _keyword in constants.attribution_analysis_keywords):
                    intent_flag = "intent_attributionAnalysis"
                    raw_intent = parser.dict2intent(text=query, disambiguation_body_list=query_body.disambiguation_body_list,
                                                    cut_tokens=cut_tokens, mode="attribution")

                    if query_body.inherited_state:
                        if pre_status:
                            pre_intent = pre_status["intent"]
                        else:
                            logger.info(f"-----------------------------------------session_id不在状态机中-------------------------------------------------")
                            pre_intent = None
                    else:
                        pre_intent = None
                    intent = parser.attribution_analysis_post_processing(query=query, intent=raw_intent, pre_intent=pre_intent)
                else:
                    raw_intent = parser.dict2intent(text=query, disambiguation_body_list=query_body.disambiguation_body_list, cut_tokens=cut_tokens)
                    intent, cur_dst = parser.dialog_management.multi_turn_intent(
                        intent=raw_intent,
                        user_id=query_body.session_id,
                        query=query,
                        tokens=parser._nl4dv.tokens,
                        token_pos_list=parser._nl4dv.token_pos_list,
                        dialog_status=dialog_status,
                        model_id=model_id,
                        pre_round_dst=pre_status_lst,
                        inherited_state=query_body.inherited_state,
                        details_flag=parser._nl4dv.details_flag
                    )
            if intent:
                logger.info("--------------------------------------intent内容--------------------------------------")
                logger.info(f"提取的Intent信息：{intent.log}")
                logger.info("-------------------------------------------------------------------------------------")
        with PerformanceTimer() as redis_timer2:
            if cur_dst:
                await RedisHelper.add_dst_to_dialog_status(
                    session_id=query_body.session_id,
                    dialog_status=dialog_status,
                    dst=cur_dst
                )
        logger.info("******************************************************************************************")
        logger.info("--------------------------------------intent请求结束--------------------------------------")
        logger.info("******************************************************************************************")
        logger.info(f"**解析耗时 {time.perf_counter() - parse_start_time:.3f} (intent {intent_timer.elapsed_time:.3f}, 切词 {cut_timer.elapsed_time:.3f}, redis {redis_timer1.elapsed_time + redis_timer2.elapsed_time:.3f}, 锁 {lock_time:.3f})")
        if not intent:
            logger.error(f"Empty intent. Failed to parse query '{query_body.sentence.query}' in subject '{query_body.sentence.subjectId}'")
            return empty_intent(query_body.sentence.query, query_body.sentence.subjectId)
        return {"successful": True, "result": {"model_id": model_id, "intent": intent, "status": intent_flag}}
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.error(f"错误的subject_id：{subject_id}，错误的model_id：{model_id} 错误的query：{query_body.sentence.query}")
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return failed_to_parse(query_body.sentence.query, query_body.sentence.subjectId)


@app.post("/api/v1/intents")
async def queries_to_intent(query_body: QueryBodys):
    results = []
    subject_id = ""
    try:
        logger.info("******************************************************************************************")
        logger.info("--------------------------------------intents请求开始--------------------------------------")
        logger.info("******************************************************************************************")
        if not valid_api_key(query_body.api_key):
            logger.error("服务器API_KEY错误！！！")
            return INVALID_API_KEY

        session_info = await RedisHelper.get_session_info_by_session_id(
            session_id=query_body.session_id,
            redis_utils=redis_utils,
            session_id2info=session_id2info
        )
        if session_info:
            subject_id = session_info["subject_id"]
            model_infos = await RedisHelper.get_model_info_by_subject_id(
                subject_id=subject_id,
                persist_table_parser_map=persist_table_parser_map,
                table_parser_map=table_parser_map,
                redis_utils=redis_utils
            )
            if not model_infos:
                logger.error(f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/")
                return not_found_parser(subject_id)
        else:
            return not_found_session(query_body.session_id)

        async with lock_query_to_intent:
            for query in query_body.query_list:
                table_id2table_name = {}  # 转换一下tableId到table_name
                table_id2model_id = {}
                model_info = {}

                model_id = query.modelId
                if not model_id:
                    parser, model_id, hits = select_model(model_infos, query.query)
                    for _model in model_infos:
                        if _model["model_id"] == model_id:
                            if not parser:
                                parser = _model["parser"]
                            model_info = _model
                            table_id2model_id = _model["table_id2model_id"]
                            for name, id in _model["table_name2table_id"].items():
                                table_id2table_name[id] = name
                            break
                else:
                    parser = None
                    for _model in model_infos:
                        if _model["model_id"] == model_id:
                            if not parser:
                                parser = _model["parser"]
                            model_info = _model
                            table_id2model_id = _model["table_id2model_id"]
                            for name, id in _model["table_name2table_id"].items():
                                table_id2table_name[id] = name
                            break
                logger.info(f"***************************************选择的model_id是{model_id}***************************************************")
                # parser = table_parser_map.get(query_body.subject_id)
                if not parser:
                    logger.error(f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/")
                    return not_found_parser(subject_id)

                # 去除query中的一些空格
                query = truncate_query(query)
                query_str = remove_spaces(query.query)

                if not query.tokens or not all(_token.type != "ORIGIN" for _token in query.tokens):
                    sentence = cut_tool.cut_query_single_model(model_info, query.tokens, query.query, [])
                    query.tokens = [TokenBody(**token) for token in sentence]
                    logger.info("------------------------------存在 ORIGIN token 调用切词-----------------------------------")

                # dict2intent_flag = ''.join([token.text for token in query.tokens]) == query.query

                for token in query.tokens:
                    if token.similarList and token.type != "UNK":
                        token.similarList = [e for e in token.similarList if e["tableId"] in table_id2model_id]
                        token.similarList = [TokenBody(**e) for e in token.similarList if table_id2model_id[e["tableId"]] == model_id]
                        for _sim in token.similarList:
                            _sim.tableId = table_id2table_name[_sim.tableId]
                    if token.tableId and token.tableId in table_id2table_name:
                        token.tableId = table_id2table_name[token.tableId]

                intent_rejection_result = model_hub.intent_reject_model(query_str)[0]
                if intent_rejection_result["label"] == "reject" and intent_rejection_result["score"] > 0.92:
                    logger.info("--------------------------------------intent_rejection-----------------------------------------")
                    results.append({"intent": None, "status": "intent_rejection", "model_id": model_id})
                else:
                    # remaining_info = extract_remaining_info_from_query(model_info=model_info, query=query_str, disambiguation_body_list=[])
                    intent = parser.dict2intent(text=query_str, disambiguation_body_list=None, cut_tokens=query.tokens)
                    logger.info("--------------------------------------intent内容--------------------------------------")
                    logger.info(f"提取的Intent信息：{intent.log if intent else intent}")
                    logger.info("-------------------------------------------------------------------------------------")
                    results.append({"intent": intent, "status": "intent_pass", "model_id": model_id})
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.error(f"错误的表格id：{subject_id}， 错误的query：{query}")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        intent = None
        results.append({"intent": intent, "status": "intent_pass"})
    # results.append(intent)

    logger.info("******************************************************************************************")
    logger.info("--------------------------------------intents请求结束--------------------------------------")
    logger.info("******************************************************************************************")
    return {"successful": True, "results": results}

@app.post("/api/v1/modify/intent")
async def modify_intent(modify_body: ModifyIntentionBody):
    logger.info("******************************************************************************************")
    logger.info("--------------------------------------modify_intent请求开始--------------------------------------")
    logger.info("******************************************************************************************")

    try:
        reject_fields = modify_body.reject_fields
        intent = modify_body.intent
        if not reject_fields:
            return {"successful": True, "intent": intent}
        else:
            subject_id = reject_fields[0].subjectId
            model_id = reject_fields[0].modelId

            model_infos = await RedisHelper.get_model_info_by_subject_id(
                subject_id=subject_id,
                persist_table_parser_map=persist_table_parser_map,
                table_parser_map=table_parser_map,
                redis_utils=redis_utils
            )
            if not model_infos:
                logger.error(f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/")
                return not_found_parser(subject_id)

            table_id2table_name = {}
            for _model in model_infos:
                if _model["model_id"] == model_id:
                    for name, id in _model["table_name2table_id"].items():
                        table_id2table_name[id] = name
                    break

            reject_fields_filter = []
            for reject_field in reject_fields:
                if reject_field.tableId in table_id2table_name:
                    reject_field.tableId = table_id2table_name[reject_field.tableId]
                    reject_fields_filter.append(reject_field)

            intent = modify_intent_by_reject_fields(intent=intent, reject_fields=reject_fields_filter)

    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return {"successful": False, "message": "something error"}

    logger.info("******************************************************************************************")
    logger.info("--------------------------------------modify_intent请求结束--------------------------------------")
    logger.info("******************************************************************************************")
    return {"successful": True, "intent": intent}

@app.post("/api/v1/sentences/compare")
def compare_sentences(sentence_compare_body: SentenceCompareBody):
    logger.info("--------------------------------------compare请求开始--------------------------------------")
    try:
        status = True
        sentence = sentence_compare_body.sentence
        compare_sentences = sentence_compare_body.compare_sentences
        _metric_info_by_sent = [token.tableId + token.tableField for token in sentence.tokens if token.type == "METRIC"]
        metric_set_by_sent = set(_metric_info_by_sent) if _metric_info_by_sent else set()
        results = []
        query2id = {}
        queries = []
        for _sent_info in compare_sentences:
            id = _sent_info.id
            _sents = _sent_info.sentences
            for _sent in _sents:
                _metric_info_by_compare_sent = [token.tableId + token.tableField for token in _sent.tokens if token.type == "METRIC"]
                metric_set_by_compare_sent = set(_metric_info_by_sent) if _metric_info_by_sent else set()
                if metric_set_by_compare_sent == metric_set_by_sent:
                    _k = "".join([token.text for token in _sent.tokens])
                    queries.append(_k)
                    query2id[_k] = id

        if queries:
            query_recalls = query_recall.recall(query="".join([token.text for token in sentence.tokens]), query_list=queries, sim_score=6.0)
            if query_recalls:
                results.append(query2id[query_recalls[0]])

    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        results = []
        status = False
    logger.info("--------------------------------------compare请求结束--------------------------------------")
    return {"successful": status, "match_question": results}

@app.post("/api/v1/retrieve/questions")
async def query_retrieve(retrieve_querybody: RetrieveQueryBody):
    logger.info("--------------------------------------retrieve请求开始--------------------------------------")
    logger.info(f"***************************传入的query是：{retrieve_querybody.query.query}**************************************************")
    # logger.info(f"***************************传入的映射问题是：{retrieve_querybody.domain_questions}**************************************************")
    try:
        if any(token.type == "ORIGIN" for token in retrieve_querybody.query.tokens):
            session_info = await RedisHelper.get_session_info_by_session_id(
                session_id=retrieve_querybody.session_id,
                redis_utils=redis_utils,
                session_id2info=session_id2info
            )
            if not session_info:
                logger.info(f"***************************session_id不在内存中**************************************************")
                return {"successful": False, "retrieveQuestions": None}
            else:
                subject_id = session_info["subject_id"]
                model_infos = await RedisHelper.get_model_info_by_subject_id(
                    subject_id=subject_id,
                    persist_table_parser_map=persist_table_parser_map,
                    table_parser_map=table_parser_map,
                    redis_utils=redis_utils
                )
                if not model_infos:
                    logger.error(f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/")
                    return {"successful": False, "retrieveQuestions": None}

                sentence, model_id = cut_tool.cut_query_multi_model(model_infos, retrieve_querybody.query.tokens, retrieve_querybody.query.query, [], "")
                retrieve_querybody.query.tokens = [TokenBody(**sub_sent) for sub_sent in sentence]

        mapping_domain_questions_infos, query_infos = domain_question_handle(retrieve_querybody.query, retrieve_querybody.domain_questions)
        if mapping_domain_questions_infos:
            results = []
            mapping_domain_questions = list(mapping_domain_questions_infos.keys())
            # retrieve_query_list = [question.query for question in domain_questions]
            query_recalls = query_recall.recall(query=retrieve_querybody.query.query, query_list=mapping_domain_questions)
            if query_recalls:
                query_index = [mapping_domain_questions_infos[query] for query in query_recalls]
                similar_infos = retrieve_querybody.domain_questions[query_index[0]]
                _, query_replace = filter_question(similar_infos.domain_question, replace_info=query_infos)
                similar_infos.domain_question = query_replace
                results.append(similar_infos)
        else:
            results = []
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.error(f"错误的query：{retrieve_querybody.query.query}")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        results = []
    logger.info("--------------------------------------retrieve请求结束--------------------------------------")
    return {"successful": True, "retrieveQuestions": results}

@app.post("/api/v1/predict")
def time_series_predict(predict_body: TimeSeriesPredictBody):
    logger.info("------------------------------------------------------进入预测--------------------------------------------------")
    try:
        ds_data = predict_body.train_data[0]["data"]
        ds_data = [datetime.utcfromtimestamp(int(ds[:-3])) for ds in ds_data]
        metric_data = predict_body.train_data[1]["data"]
        logger.info(f"训练数据是：{predict_body}")
        fields_name = [element["fieldName"] for element in predict_body.train_data]
        fields_type = [element["fieldType"] for element in predict_body.train_data]
        fields = [Field(name, type) for name, type in zip(fields_name, fields_type)]
        train_data = [[cur_ds_data, cur_metric_data] for (cur_ds_data, cur_metric_data) in zip(ds_data, metric_data)]

        # delta_time = (datetime.utcfromtimestamp(int(predict_body.train_data[0]["endDate"][:-3])) - datetime.utcfromtimestamp(int(predict_body.train_data[0]["startDate"][:-3]))).days
        time_unit = predict_body.train_data[0]["timeUnit"][:1]
        if time_unit in ["D", "M", "W"]:
            window_size = 100
        elif time_unit == "Q":
            window_size = 100
        else:
            window_size = 100

        logger.info(f"--------------------------------预测的时间窗口是{window_size}---------------------------------------------")

        df = pd.DataFrame(train_data, columns=['ds', 'y'])
        df['ds'] = pd.to_datetime(df['ds'])
        log_flag = False
        if any(ele <= 0 for ele in df["y"]):
            log_flag = False
        if log_flag:
            df['y'] = np.log(df['y'])
        df['cap'] = max(abs(df["y"])) * 1.2
        df['floor'] = min(abs(df["y"])) * 1.1

        model = model_hub._init_time_series_prediction()
        # model.add_seasonality(name='monthly', period=30.5, fourier_order=3)
        model.fit(df)

        # 创建未来日期的DataFrame，用于预测
        future = model.make_future_dataframe(periods=window_size, freq=time_unit)
        future['cap'] = max(abs(df["y"])) * 1.2
        future['floor'] = min(abs(df["y"])) * 1.1

        forecast = model.predict(future)
        tail_forecast = forecast[['ds', 'yhat', 'yhat_lower', 'yhat_upper']].tail(window_size)
        if log_flag:
            forecast_yhat = list(np.exp(tail_forecast['yhat']))
            forecast_yhay_lower = list(np.exp(tail_forecast['yhat_lower']))
            forecast_yhay_upper = list(np.exp(tail_forecast['yhat_upper']))
        else:
            forecast_yhat = list(tail_forecast['yhat'])
            forecast_yhay_lower = list(tail_forecast['yhat_lower'])
            forecast_yhay_upper = list(tail_forecast['yhat_upper'])
        forecast_ds = [str(int(ds.timestamp()) * 1000) for ds in tail_forecast["ds"]]
        forecast_filter = [(ds, yhat, yhat_lower, yhat_upper) for (ds, yhat, yhat_lower, yhat_upper) in zip(forecast_ds, forecast_yhat, forecast_yhay_lower, forecast_yhay_upper)]
        results = {
            "fields": fields,
            "points": forecast_filter
        }
        del model
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        results = {}
    return {"successful": True, "forecast": results}

@app.post("/api/v1/anomaly/detect")
def anomaly_detection(anomaly_detection_body: AnomalyDetectionBody):
    logger.info("------------------------------------------------------异常检测--------------------------------------------------")
    try:
        anomaly_results = []
        detect_util = AnomalyDetection()

        # 如果异常检测失败，则返回空数组
        date_dict_list = [anomaly_detection_body.train_data[i] for i in range(len(anomaly_detection_body.train_data)) if anomaly_detection_body.train_data[i]["fieldType"] == "Date"]
        for date_dict in date_dict_list:
            if any((not date or not date.isdigit()) for date in date_dict["data"]):
                return {"successful": True, "anomaly_results": anomaly_results}

        metric_datas = [anomaly_detection_body.train_data[i] for i in range(len(anomaly_detection_body.train_data)) if anomaly_detection_body.train_data[i]["fieldType"] == "Metric"]
        metric_dict = {}
        for data in metric_datas:
            idx_map = {}
            _metric_data = []
            for idx, _data in enumerate(data["data"]):
                if isinstance(_data, float) or isinstance(_data, int):
                    if not idx_map:
                        idx_map["0"] = idx
                    else:
                        idx_map[str(len(idx_map))] = idx
                    _metric_data.append(_data)
            metric_dict[data["fieldName"]] = {
                "data": _metric_data,
                "idx_map": idx_map
            }
        for field_name, metric_info in metric_dict.items():
            metric_data = metric_info["data"]
            idx_map = metric_info["idx_map"]
            metric_data = np.array(metric_data)
            anomaly_result = detect_util.detect(metrics=metric_data, fieldName=field_name, idx_map=idx_map)
            anomaly_results.append(anomaly_result)

    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        anomaly_results = []
    return {"successful": True, "anomaly_results": anomaly_results}

@app.post("/api/v1/retrieval/data")
async def retrieval_logic(retrieval_logic_body: RetrievalLogicBody):
    try:
        subject_id = retrieval_logic_body.subject_id
        model_infos = await RedisHelper.get_model_info_by_subject_id(
            subject_id=subject_id,
            persist_table_parser_map=persist_table_parser_map,
            table_parser_map=table_parser_map,
            redis_utils=redis_utils
        )
        if not model_infos:
            logger.error(f"Found no Parser for subject_id '{subject_id}'. Please upload the table schema first via POST /api/v1/datasets/")
            return not_found_parser(subject_id)
        model_id = retrieval_logic_body.model_id
        parser = None
        for _model in model_infos:
            if _model["model_id"] == model_id:
                parser = _model["parser"]
                break

        logger.info(f"***************************************选择的model_id是{model_id}***************************************************")
        if not parser:
            return not_found_parser(subject_id)
        historical_intent = parser.parse_furture_info(retrieval_logic_body.intent)
        logger.info(f"-----------------------------------------------取数的时间范围是{historical_intent['date_range']['from_date']}到{historical_intent['date_range']['to_date']}--------------------------------------------------")
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        historical_intent = None
    return {"successful": True, "historicalIntent": historical_intent}


@app.get("/api/v1/alive")
async def is_alive():
    return "success"


@app.get("/error")
async def raise_error():
    raise Exception("error")


@app.get("/api/v1/exist/subject")
async def get_exist_subject():
    try:
        exist_keys = await RedisHelper.get_exist_persist_keys(
            persist_table_parser_map=persist_table_parser_map,
            redis_utils=redis_utils
        )
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        exist_keys = []
    return {"subject_ids": exist_keys}

@app.get("/api/v1/version/model")
async def get_version():
    return {
        "version": VERSION
    }

@app.post("/api/v1/upload_openai_key")
def upload_key(openai_key_body: OpenaiKeyBody):
    cur_key = openai_key_body.key
    file_path = OPENAI_API_KEY_PATH
    try:
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(
                json.dumps({"key": cur_key})
            )
        logger.info("重写openai_key")
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        return {"successful": False, "message": "文件存储失败"}
    return {"successful": True, "message": "key导入成功"}


@app.get("/api/v1/get_openai_key")
def get_key():
    file_path = OPENAI_API_KEY_PATH
    if not os.path.exists(file_path):
        return {"key": "", "message": "key文件为空"}
    else:
        with open(file_path, "r", encoding="utf-8") as f:
            config_file = json.load(f)
        key = config_file.get("key", "")
        return {"key": key, "message": "key读取成功"}

@app.get("/api/v1/session/dialog/status/{session_id}/{index}")
async def get_dialog_status_by_session(session_id: str, index: int):
    try:
        pre_status_lst = await RedisHelper.get_pre_status_by_session_id(
            session_id=session_id,
            dialog_status=dialog_status
        )
        result = None
        if pre_status_lst and index <= len(pre_status_lst):
            result = pre_status_lst[index-1]
    except Exception as e:
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！错误信息！！！！！！！！！！！！！！！！！！！！！！！！！！")
        logger.exception(e)
        logger.info("！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！！")
        result = None

    if not result:
        return {"successful": False, "dialog_status": None, "message": f"{session_id} not in dialog_status"}
    else:
        return {"successful": True, "dialog_status": result, "message": "session in dialog_status"}


@app.put("/loglevel/{level}")
async def set_log_level(level: str):
    try:
        level = level.strip().lower()
        if level == "critical":
            log_level = logging.CRITICAL
        elif level == "fatal":
            log_level = logging.FATAL
        elif level == "error":
            log_level = logging.ERROR
        elif level == "warning":
            log_level = logging.WARNING
        elif level == "info":
            log_level = logging.INFO
        elif level == "debug":
            log_level = logging.DEBUG
        else:
            logger.warning(f"未知的日志级别 '{level}'")
            return False
        logger.setLevel(log_level)
    except Exception as e:
        logger.exception(e)
        logger.info("调整日志级别时出现未知错误")
        return False
    return True


# 如果计算机设备有连接到 Internet 的路由，可以使用以下方法获取局域网ip
def get_better_local_ip():
    skt = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    skt.connect(("8.8.8.8", 80))
    return skt.getsockname()[0]

if __name__ == "__main__":
    # 下行代码获取的ip地址有可能会得到虚拟机的ip地址，若不好用可尝试 get_better_local_ip 函数
    # ip = socket.gethostbyname(socket.getfqdn(socket.gethostname()))

    # 如果计算机设备有连接到 Internet 的路由，可以使用以下
    ip = get_better_local_ip()

    uvicorn.run(app='app:app', host=ip, port=PORT, reload=False)
